

set -ex



cd tests
python -c "import daal4py"
mpirun -n 4 python -m unittest discover -v -p spmd*.py
python -m unittest discover -v -p 'test*.py'
pytest --pyargs ../daal4py/sklearn/
pytest --pyargs ../onedal/
python ../examples/daal4py/run_examples.py
python -m daal4py ../examples/daal4py/sycl/sklearn_sycl.py
exit 0
